{****************************************************************************
** Demonstration of automatically detecting the Sound Blaster settings     **
**  by Steven H Don                                                        **
**                                                                         **
** For questions, feel free to e-mail me.                                  **
**                                                                         **
**    shd@earthling.net                                                    **
**    http://shd.cjb.net                                                   **
**                                                                         **
****************************************************************************}
program SBAutoDetect;

uses Dos, Crt;

var
  Base              : Word;    {Sound Blaster base address}
  BPort, CType      : String;  {Base port as string type and card}
  LoDMA, HiDMA, IRQ : Byte;    {DMA Channels and IRQ}
  Old71, Old11,
  Old13, Old15,
  Old72, Old73      : Pointer; {Pointers to old interrupt routines}

{****************************************************************************
** Checks to see if a Sound Blaster exists at a given address, returns     **
** true if Sound Blaster found, false if not.                              **
****************************************************************************}
function ResetDSP(Test : Word) : Boolean;
begin
  {Reset the DSP}
  Port[Test + $6] := 1;
  Delay(10);
  Port[Test + $6] := 0;
  Delay(10);
  {Check if reset was succesfull}
  if (Port[Test + $E] and $80 = $80) and (Port[Test + $A] = $AA) then begin
    {DSP was found}
    ResetDSP := true;
    Base := Test;
  end else
    {No DSP was found}
    ResetDSP := false;
end;

{****************************************************************************
** Read a byte from the DSP (Digital Signal Processor) on the Sound Blaster**
****************************************************************************}
function ReadDSP : byte;
begin
  while Port[Base + $E] and $80 = 0 do;
  ReadDSP := Port[Base + $A];
end;

{****************************************************************************
** Send a byte to the DSP (Digital Signal Processor) on the Sound Blaster  **
****************************************************************************}
procedure WriteDSP(Value : byte);
begin
  {Wait for the DSP to be ready to accept data}
  while Port[Base + $C] and $80 <> 0 do;
  {Send byte}
  Port[Base + $C] := value;
end;

{****************************************************************************
** Returns the version number of the DSP.                                  **
****************************************************************************}
function DSPVersion : Word;
var
  Version : Word;

begin
  WriteDSP ($E1);
  Version := ReadDSP;
  DSPVersion := ReadDSP + Version shl 8;
end;

{****************************************************************************
** This tells the DSP to start playing a sample. However, the DMA          **
** controller has not been set up. Therefore the DSP signals the DMA       **
** controller for data. We check to see where that call takes place and    **
** thus determine the DMA channel the DSP uses                             **
****************************************************************************}
function TestLoDMA : Byte;
var
  Before, After : Byte;
  Count : Word;

begin
  {Reset DSP}
  ResetDSP (Base);
  WriteDSP($D3);             {DSP-command D3h - Speaker Off}
  WriteDSP($40);             {DSP-command 40h - Set sample frequency}
  WriteDSP(165);             {Write time constant for 11025Hz}
  WriteDSP($14);             {DSP-command 14h - 8bit playback}
  WriteDSP($03);             {Set the block length to 3 = 1 DWORD}

  Count := 0;

  {Disable interrupts}
  asm
    cli
  end;

  {Look at the current state of the DMA requests}
  Before := Port [8] and $E0;

  {Start playback, * note : this is still part of "set block length"}
  WriteDSP($00);

  repeat
    {Check them again until there is a change or it takes too long}
    After := Port [8] and $E0;
    inc (Count);
  until (Before <> After) or (Count > 60000);

  {Reenable interrupts}
  asm
    sti
  end;

  {Reset DSP}
  ResetDSP (Base);

  {Is there a change?}
  if Before <> After then begin
    Count := 0;
    After := After xor Before;
    {If yes, which bit is set?}
    while (After > 0) do begin
      inc (Count);
      After := After shr 1;
    end;
    TestLoDMA := Count - 5;
    Exit;
  end;
  {No DMA???}
  TestLoDMA := $FF;
end;

{****************************************************************************
** This does the same thing for the 16bit DMA channel, if used             **
****************************************************************************}
function TestHiDMA : Byte;
var
  Before, After : Byte;
  Count : Word;

begin
  {Reset DSP}
  ResetDSP (Base);
  WriteDSP ($B0);            {DSP-command B0h - 16bit DMA}
  WriteDSP ($10);            {Mono}
  WriteDSP ($03);            {Set the block length to 3 = 1 DWORD}

  Count := 0;

  {Disable interrupts}
  asm
    cli
  end;

  {Look at the current state of the DMA requests}
  Before := Port [$D0] and $F0;

  {Start playback, * note : this is still part of "set block length"}
  WriteDSP($00);

  repeat
    {Check them again until there is a change or it takes too long}
    After := Port [$D0] and $F0;
    inc (Count);
  until (Before <> After) or (Count > 60000);

  {Reenable interrupts}
  asm
    sti
  end;

  {Reset DSP}
  ResetDSP (Base);

  {Is there a change?}
  if Before <> After then begin
    Count := 0;
    After := After xor Before;
    {If yes, which bit is set?}
    while (After > 0) do begin
      inc (Count);
      After := After shr 1;
    end;
    TestHiDMA := Count - 1;
    Exit;
  end;
  {No DMA???}
  TestHiDMA := $FF;
end;

{****************************************************************************
** This tells the DSP to start playing a sample. And waits to see which    **
** one of the IRQ service routines is called.                              **
****************************************************************************}
procedure TestIRQ;
var
  Count : Word;

begin
  IRQ := 0;
  WriteDSP($D1);                {DSP-command D1h - Enable speaker, required
                                on older SB's}
  WriteDSP($40);                {DSP-command 40h - Set sample frequency}
  WriteDSP(165);                {Write time constant for 11025Hz}
  Port[$0A] := 4 + LoDMA;       {Mask DMA channel}
  Port[$0C] := 0;               {Clear byte pointer}
  Port[$0B] := $49;             {Set mode}

  Port[1 + LoDMA shl 1] := $01; {Set the block length to 1 = 1 byte}
  Port[1 + LoDMA shl 1] := $00;

  Port[$0A] := LoDMA;           {Unmask DMA channel}

  WriteDSP($14);                {DSP-command 14h - Single cycle playback}
  WriteDSP($01);
  WriteDSP($00);

  Count := 0;
  repeat
    {Wait until either one of the IRQ service routines has been called
    or it takes too long}
    inc (Count);
  until (IRQ <> 0) or (Count = $FFFF);

end;

{****************************************************************************
** IRQ service routine - this is called when the DSP has finished playing  **
** a block.                                                                **
****************************************************************************}
procedure ServiceIRQ2; interrupt;
var
  Temp : Byte;

begin
  {Relieve DSP}
  Temp := Port[Base + $E];
  {Acknowledge hardware interrupt}
  Port[$20] := $20;
  {Since IRQ 2 has been redirected, we need to acknowledge two interrupt
  controllers}
  Port[$A0] := $20;
  {Save interrupt channel number}
  IRQ := 2;
end;

procedure ServiceIRQ3; interrupt;
var
  Temp : Byte;

begin
  Temp := Port[Base + $E]; Port[$20] := $20; IRQ := 3;
end;

procedure ServiceIRQ5; interrupt;
var
  Temp : Byte;

begin
  Temp := Port[Base + $E]; Port[$20] := $20; IRQ := 5;
end;

procedure ServiceIRQ7; interrupt;
var
  Temp : Byte;

begin
  Temp := Port[Base + $E]; Port[$20] := $20; IRQ := 7;
end;

procedure ServiceIRQ10; interrupt;
var
  Temp : Byte;

begin
  Temp := Port[Base + $E]; Port[$20] := $20; Port[$A0] := $20; IRQ := 10;
end;

procedure ServiceIRQ11; interrupt;
var
  Temp : Byte;

begin
  Temp := Port[Base + $E]; Port[$20] := $20; Port[$A0] := $20; IRQ := 11;
end;

var
  Temp : Byte;

begin
  {Clear screen}
  ClrScr;
  {Check for Sound Blaster, address: ports 220, 230, 240, 250, 260 or 280}
  for Temp := 1 to 8 do begin
    if Temp <> 7 then
    if ResetDSP ($200 + Temp shl 4) then begin
      BPort := '200h';
      inc (BPort [2], Temp);
      Break
    end;
  end;
  if Temp = 9 then begin
    {or none at all}
    Writeln ('No Sound Blaster found');
    Halt;
  end;

  {There are several different type of Sound Blaster}
  if DSPVersion = $100 then CType := 'Sound Blaster';
  if DSPVersion = $105 then CType := 'Sound Blaster 1.5';
  if DSPVersion = $200 then CType := 'Sound Blaster Pro 2';
  if DSPVersion = $300 then CType := 'Sound Blaster Pro 3';
  if hi (DSPVersion) >= 4 then CType := 'Sound Blaster 16/ASP/AWE 32/AWE 64';

  {Check for DMA Channels}
  LoDMA := TestLoDMA;
  if hi (DSPVersion) >= 4 then begin
    HiDMA := TestHiDMA;
    if HiDMA = $FF then HiDMA := LoDMA;
  end;

  {Save old IRQ vectors. All the standard possible Sound Blaster IRQs.
  An SB can be at IRQ 2, 3, 5, 7 or 10. Some clones can also use 11.
  IRQ 2 has been redirected to IRQ 9 (int $71)}
  GetIntVec($71, Old71);
  GetIntVec(11, Old11);
  GetIntVec(13, Old13);
  GetIntVec(15, Old15);
  GetIntVec($72, Old72);
  GetIntVec($73, Old73);

  {Set new IRQ vectors}
  SetIntVec($71, Addr (ServiceIRQ2));
  SetIntVec(11, Addr (ServiceIRQ3));
  SetIntVec(13, Addr (ServiceIRQ5));
  SetIntVec(15, Addr (ServiceIRQ7));
  SetIntVec($72, Addr (ServiceIRQ10));
  SetIntVec($73, Addr (ServiceIRQ11));

  {Enable IRQs}
  Port[$21]:=Port[$21] and not (1 shl 2);
  Port[$21]:=Port[$21] and not (1 shl 3);
  Port[$21]:=Port[$21] and not (1 shl 5);
  Port[$21]:=Port[$21] and not (1 shl 7);
  Port[$A1]:=Port[$A1] and not (1 shl (9 - 8));
  Port[$A1]:=Port[$A1] and not (1 shl (10 - 8));
  Port[$A1]:=Port[$A1] and not (1 shl (11 - 8));

  {Scan which IRQ is called}
  TestIRQ;

  {Free interrupt vectors used to service IRQs}
  SetIntVec($71, Old71);
  SetIntVec(11, Old11);
  SetIntVec(13, Old13);
  SetIntVec(15, Old15);
  SetIntVec($72, Old72);
  SetIntVec($73, Old73);

  {Mask IRQs}
  Port[$21]:=Port[$21] or (1 shl 2);
  Port[$21]:=Port[$21] or (1 shl 3);
  Port[$21]:=Port[$21] or (1 shl 5);
  Port[$21]:=Port[$21] or (1 shl 7);
  Port[$A1]:=Port[$A1] or (1 shl (9 - 8));
  Port[$A1]:=Port[$A1] or (1 shl (10 - 8));
  Port[$A1]:=Port[$A1] or (1 shl (11 - 8));

  {Display results}
  Writeln ('Card                : ', CType);
  Writeln ('Base port           : ', BPort);
  Writeln ('IRQ                 : ', IRQ);
  Writeln ('DMA channel (8bit)  : ', LoDMA);
  if hi (DSPVersion) >= 4 then
    Writeln ('DMA channel (16bit) : ', HiDMA);
end.